"""Initialize from the environment"""
import logging
import logging.config
import os


def enabled():
    return os.getenv("APPMAP", "false") == "true"


def _configure_logging():
    log_level = os.getenv("APPMAP_LOG_LEVEL", "warning").upper()

    log_config = os.getenv("APPMAP_LOG_CONFIG")
    log_stream = os.getenv("APPMAP_LOG_STREAM", "stderr")
    log_stream = f'ext://sys.{log_stream}'
    config_dict = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'style': '{',
                'format': '[{asctime}] {levelname} {name}: {message}'
            }
        },
        'handlers': {
            'default': {
                'class': 'logging.StreamHandler',
                'formatter': 'default'
            }
        },
        'loggers': {
            'root': {
                'level': log_level,
                'handlers': ['default']
            }
        }
    }
    if log_config is not None:
        name, level = log_config.split('=', 2)
        config_dict['loggers'].update({
            name: {
                'level': level.upper(),
                'handlers': ['default'],
                'propagate': False
            }
        })
    logging.config.dictConfig(config_dict)


def initialize():
    _configure_logging()
    logging.info('appmap enabled: %s', enabled())



import pytest
from appmap._implementation.recording import Recorder

pytest.importorskip('django')


def test_sql_capture():
    # pylint: disable=unused-import
    import django.conf
    import django.db
    import appmap.django  # noqa: F401

    recorder = Recorder()
    recorder.events().clear()
    recorder.enabled = True

    django.conf.settings.configure(DATABASES={'default': {'ENGINE': 'django.db.backends.sqlite3', 'NAME': ':memory:'}})

    conn = django.db.connections['default']
    conn.cursor().execute('SELECT 1').fetchall()

    recorder.enabled = False
    assert recorder.events()[0].sql_query['sql'] == 'SELECT 1'



"""AppMap recorder for Python"""

from ._implementation.env import enabled # noqa; F401
from ._implementation.recording import Recording  # noqa: F401
from ._implementation import generation  # noqa: F401



from . import configuration
from . import env
from . import event
from . import recording
from .py_version_check import check_py_version


def initialize():
    check_py_version()
    env.initialize()
    event.initialize()
    recording.initialize()
    configuration.initialize()  # needs to be initialized after recording


initialize()





class ClassMethodMixin:
    @classmethod
    def class_method(cls):
        return f'ClassMethodMixin#class_method, cls {cls}'


class Super:
    def instance_method(self):
        return 'Super#instance_method'


class ExampleClass(Super, ClassMethodMixin):
    def __repr__(self):
        return f'ExampleClass and {self.another_method()}'

    @staticmethod
    def static_method():
        return 'ExampleClass.static_method'

    def another_method(self):
        return "ExampleClass#another_method"

    def test_exception(self):
        raise Exception('test exception')



import platform
import re

from .utils import subprocess_run


class Metadata:
    @staticmethod
    def to_dict():
        metadata = {
            'language': {
                'name': 'python',
                'engine': platform.python_implementation(),
                'version': platform.python_version()
            },
            'client': {
                'name': 'appmap',
                'url': 'https://github.com/applandinc/appmap-python'
            }
        }

        if Metadata._git_available():
            metadata.update({'git': Metadata._git_metadata()})

        return metadata

    @staticmethod
    def _git_available():
        ret = subprocess_run(['git', 'status'])
        if not ret.returncode:
            return True

        return False

    @staticmethod
    def _git_metadata():
        git_repo = subprocess_run(['git', 'config', '--get', 'remote.origin.url']).stdout.strip()
        git_branch = subprocess_run(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).stdout.strip()
        git_sha = subprocess_run(['git', 'rev-parse', 'HEAD']).stdout.strip()
        git_status = list(map(lambda x: x.strip(), subprocess_run(['git', 'status', '-s']).stdout.strip().split('\n')))
        git_last_annotated_tag = subprocess_run(['git', 'describe', '--abbrev=0']).stdout.strip() or None
        git_last_tag = subprocess_run(['git', 'describe', '--abbrev=0', '--tags']).stdout.strip() or None

        pattern = re.compile(r'.*-(\d+)-\w+$')

        git_commits_since_last_annotated_tag = 0
        if git_last_annotated_tag:
            result = pattern.search(subprocess_run(['git', 'describe']).stdout.strip())
            if result:
                git_commits_since_last_annotated_tag = int(result[1])

        git_commits_since_last_tag = 0
        if git_last_tag:
            result = pattern.search(subprocess_run(['git', 'describe', '--tags']).stdout.strip())
            if result:
                git_commits_since_last_tag = int(result[1])

        return {
            'repository': git_repo,
            'branch': git_branch,
            'commit': git_sha,
            'status': git_status,
            'git_last_annotated_tag': git_last_annotated_tag,
            'git_last_tag': git_last_tag,
            'git_commits_since_last_annotated_tag': git_commits_since_last_annotated_tag,
            'git_commits_since_last_tag': git_commits_since_last_tag
        }



import platform
import sys


class AppMapPyVerException(Exception):
    pass


def check_py_version():
    MIN_PY_VERSION = (3, 9)
    if sys.version_info < MIN_PY_VERSION:
        raise AppMapPyVerException(f'Minimum Python version supported is {MIN_PY_VERSION}, found {platform.python_version()}')



"""
Test event functionality
"""
import os
import sys
from functools import partial
from queue import Queue
from threading import Thread

import appmap
import appmap._implementation
from appmap._implementation.event import _EventIds


def test_per_thread_id():
    """ thread ids should be constant for a  thread """
    assert _EventIds.get_thread_id() == _EventIds.get_thread_id()


def test_thread_ids():
    """thread ids should be different per thread"""

    tids = Queue()

    def add_thread_id(q):
        tid = _EventIds.get_thread_id()
        q.put(tid)

    t = partial(add_thread_id, tids)
    threads = [Thread(target=t) for _ in range(5)]
    list(map(Thread.start, threads))
    list(map(Thread.join, threads))

    assert not tids.empty()

    all_tids = [tids.get() for _ in range(tids.qsize())]
    assert len(set(all_tids)) == len(all_tids)  # Should all be unique


def test_recursion_protection(data_dir, monkeypatch):
    monkeypatch.setenv("APPMAP", "true")
    monkeypatch.setenv("APPMAP_CONFIG", os.path.join(data_dir, 'appmap.yml'))
    monkeypatch.setenv("APPMAP_LOG_LEVEL", "debug")

    sys.path.append(data_dir)
    appmap._implementation.initialize()  # pylint: disable=protected-access
    r = appmap.Recording()
    with r:
        from example_class import ExampleClass  # pylint: disable=import-error
        ExampleClass().instance_method()

    # If we get here, recursion protection for rendering the receiver
    # is working
    assert True



"""Generate an AppMap"""
import orjson
from dataclasses import dataclass, field

from .metadata import Metadata
from .event import Event


class ClassMapDict(dict):
    pass


@dataclass
class ClassMapEntry:
    name: str
    type: str


@dataclass
class PackageEntry(ClassMapEntry):
    children: ClassMapDict = field(default_factory=ClassMapDict)
    type: str = 'package'


@dataclass
class ClassEntry(ClassMapEntry):
    children: ClassMapDict = field(default_factory=ClassMapDict)
    type: str = 'class'


@dataclass
class FuncEntry(ClassMapEntry):
    location: str
    static: bool

    def __init__(self, e):
        super().__init__(e.method_id, 'function')
        self.location = f'{e.path}:{e.lineno}'
        self.static = e.static


def classmap(recording):
    ret = ClassMapDict()
    for e in recording.events:
        try:
            if e.event != 'call':
                continue
            packages, class_ = e.defined_class.rsplit('.', 1)
            children = ret
            for p in packages.split('.'):
                entry = children.setdefault(p, PackageEntry(p))
                children = entry.children

            entry = children.setdefault(class_, ClassEntry(class_))
            children = entry.children

            loc = f'{e.path}:{e.lineno}'
            children.setdefault(loc, FuncEntry(e))
        except AttributeError:
            # Event might not have a defined_class attribute;
            # SQL events for example are calls without it.
            # Ignore them when building the class map.
            continue

    return ret


def appmap(recording, metadata):
    appmap_metadata = Metadata.to_dict()
    if metadata:
        appmap_metadata.update(metadata)

    return {
        'version': '1.4',
        'metadata': appmap_metadata,
        'events': recording.events,
        'classMap': list(classmap(recording).values())
    }


class AppMapEncoder:
    @staticmethod
    def default(o):
        if isinstance(o, Event):
            return o.to_dict()
        elif isinstance(o, ClassMapDict):
            return list(o.values())
        elif isinstance(o, ClassMapEntry):
            return vars(o)
        raise TypeError


def dump(recording, metadata=None):
    a = appmap(recording, metadata)
    return orjson.dumps(a, default=AppMapEncoder.default,
                        option=(orjson.OPT_PASSTHROUGH_SUBCLASS |
                                orjson.OPT_PASSTHROUGH_DATACLASS))





import inspect
import logging
import sys

from abc import ABC, abstractmethod
from functools import wraps

from . import env
from . import utils

logger = logging.getLogger(__name__)


class Recording:
    """
    Context manager to make it easy to capture a Recording.  exit_hook
    will be called when the block exits, before any exceptions are
    raised.
    """
    def __init__(self, exit_hook=None):
        self.events = []
        self.exit_hook = exit_hook

    def start(self):

        if not env.enabled():
            return

        Recorder().start_recording()

    def stop(self):
        if not env.enabled():
            return False

        self.events = Recorder().stop_recording()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_value, traceback):
        logger.debug("Recording.__exit__, stopping with exception %s", exc_type)
        self.stop()
        if self.exit_hook is not None:
            self.exit_hook(self)
        return False


class Filter(ABC):
    def __init__(self, next_filter):
        self.next_filter = next_filter

    @abstractmethod
    def filter(self, class_):
        """
        Determine whether the given class should have its methods
        instrumented.  Return True if it should be, False if it should
        not be, or call the next filter if this filter can't decide.
        """

    @abstractmethod
    def wrap(self, fn, isstatic):
        """
        Determine whether the given method should be instrumented.  If
        so, return a new function that wraps the old.  If not, return
        the original function.  isstatic is True if fn is a static or
        class method, False otherwise.
        """


class NullFilter(Filter):
    def __init__(self, next_filter=None):
        super().__init__(next_filter)

    def filter(self, class_):
        return False

    def wrap(self, fn, isstatic):
        return fn


def is_class(c):
    # We don't want to use inspect.isclass here. It uses isinstance to
    # check the class of the object, which will invoke any method
    # bound to __class__. (For example, celery.local.Proxy uses this
    # mechanism to return the class of the proxied object.)
    #
    # We want the *real* class of c, not whatever it wants to claim as
    # its class. type() give us that.
    return type(c) == type


def get_classes(module):
    return [v for __, v in module.__dict__.items() if is_class(v)]


def get_members(class_):
    """
    Get the function members of the given class.  Unlike
    inspect.getmembers, this function only calls getattr on functions,
    to avoid potential side effects.

    Returns a list of tuples of the form (key, dict_value, attr_value):
      * key is the attribute name
      * dict_value is class_.__dict__[key]
      * and attr_value is getattr(class_, key)
    """
    def is_member_func(m):
        return (inspect.isfunction(m)
                or inspect.ismethod(m)
                or utils.is_staticmethod(m)
                or utils.is_classmethod(m))

    ret = []
    for key in dir(class_):
        if key.startswith('__'):
            continue
        value = inspect.getattr_static(class_, key)
        if not is_member_func(value):
            continue
        ret.append((key, value))
    return ret


class Recorder:
    """ Singleton Recorder class """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            logger.debug('Creating the Recorder object')
            cls._instance = super(Recorder, cls).__new__(cls)

            # Put any __init__ here.
            cls._instance._initialized = False

        return cls._instance

    def __init__(self):
        if self.__getattribute__('_initialized'):  # keep pylint happy
            return
        self._initialized = True
        self.enabled = False
        self.filter_stack = [NullFilter]
        self.filter_chain = []
        self._events = []

    @classmethod
    def initialize(cls):
        cls._instance = None

    def use_filter(self, filter_class):
        self.filter_stack.append(filter_class)

    def start_recording(self):
        logger.debug('AppMap recording started')
        self.enabled = True

    def stop_recording(self):
        logger.debug('AppMap recording stopped')
        self.enabled = False
        return self._events

    def add_event(self, event):
        """
        Add an event to the current recording
        """
        self._events.append(event)

    def events(self):
        """
        Get the events from the current recording
        """
        return self._events

    def do_import(self, *args, **kwargs):
        mod = args[0]
        logger.debug('do_import, mod %s args %s kwargs %s', mod, args, kwargs)
        if not self.filter_chain:
            logger.debug('  filter_stack %s', self.filter_stack)
            self.filter_chain = self.filter_stack[0](None)
            for filter_ in self.filter_stack[1:]:
                self.filter_chain = filter_(self.filter_chain)
                logger.debug('  self.filter chain: %s', self.filter_chain)

        if mod.__name__.startswith('appmap'):
            return

        classes = get_classes(mod)
        logger.debug('  classes %s', classes)
        for class_ in classes:
            if not self.filter_chain.filter(class_):
                continue

            logger.debug('  looking for members')
            functions = get_members(class_)
            logger.debug('  functions %s', functions)
            for fn_name, fn in functions:
                isstatic = utils.is_staticmethod(fn)
                isclass = utils.is_classmethod(fn)

                if isstatic or isclass:
                    new_fn = self.filter_chain.wrap(fn.__func__, isstatic=True)
                else:
                    new_fn = self.filter_chain.wrap(fn, isstatic=False)

                if new_fn != fn:
                    if isstatic:
                        new_fn = staticmethod(new_fn)
                    elif isclass:
                        new_fn = classmethod(new_fn)
                    setattr(class_, fn_name, new_fn)


def wrap_exec_module(exec_module):
    @wraps(exec_module)
    def wrapped_exec_module(*args, **kwargs):
        logger.debug(('exec_module %s'
                      ' exec_module.__name__ %s'
                      ' args %s'
                      ' kwargs %s'),
                     exec_module,
                     exec_module.__name__,
                     args,
                     kwargs)
        exec_module(*args, **kwargs)
        Recorder().do_import(*args, **kwargs)
    return wrapped_exec_module


def wrap_find_spec(find_spec):
    @wraps(find_spec)
    def wrapped_find_spec(*args, **kwargs):
        spec = find_spec(*args, **kwargs)
        if spec is not None:
            if getattr(spec.loader, "exec_module", None) is not None:
                loader = spec.loader
                logger.debug("wrap_find_spec, before loader.exec_module %s",
                             loader.exec_module)
                loader.exec_module = wrap_exec_module(loader.exec_module)
                logger.debug("  after loader.exec_module %s",
                             loader.exec_module)
            else:
                logger.warning("%s doesn't have exec_module", spec.loader)
        return spec
    return wrapped_find_spec


def initialize():
    Recorder.initialize()
    if env.enabled():
        wrapped_attr = '_appmap_find_spec'
        for h in sys.meta_path:
            fn = inspect.getattr_static(h, 'find_spec', None)
            if fn is None:
                continue
            if getattr(fn, wrapped_attr, None) is not None:
                logger.debug('meta path finder find_spec, already wrapped')
                continue

            # XXX This processing of the find_spec method is very
            # similar to the processing of instrumented functions
            # above. At some point, we should see if we can DRY them
            # up.
            isstatic = utils.is_staticmethod(fn)
            isclass = utils.is_classmethod(fn)
            if isstatic or isclass:
                # Wrap the function that was decorated (by
                # staticmethod or classmethod)
                new_fn = wrap_find_spec(fn.__func__)
                if isstatic:
                    new_fn = staticmethod(new_fn)
                elif isclass:
                    new_fn = classmethod(new_fn)
                setattr(new_fn, wrapped_attr, True)
            else:
                # Wrap the function, bind it to the finder instance
                # that's on sys.meta_path.
                new_fn = wrap_find_spec(fn)
                setattr(new_fn, wrapped_attr, True)
                new_fn = new_fn.__get__(h)

            h.find_spec = new_fn



"""Test recording context manager"""
import json
import os
import sys

from .appmap_test_base import AppMapTestBase


class TestRecording(AppMapTestBase):
    def test_recording_works(self, data_dir, monkeypatch):
        with open(os.path.join(data_dir, 'expected.appmap.json')) as f:
            expected_appmap = json.load(f)

        sys.path.append(data_dir)

        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", "debug")

        import appmap
        # Reinitialize to pick up the environment variables just set
        appmap._implementation.initialize()  # pylint: disable=protected-access
        r = appmap.Recording()
        with r:
            from example_class import ExampleClass  # pylint: disable=import-error
            ExampleClass.static_method()
            ExampleClass.class_method()
            ExampleClass().instance_method()
            try:
                ExampleClass().test_exception()
            except:  # pylint: disable=bare-except  # noqa: E722
                pass

        generated_appmap = self.normalize_appmap(appmap.generation.dump(r))

        assert generated_appmap == expected_appmap



import os
import os.path

import json

from .appmap_test_base import AppMapTestBase


class TestPytest(AppMapTestBase):
    def test_basic_integration(self, data_dir, testdir):
        testdir.copy_example('pytest')
        testdir.monkeypatch.setenv('APPMAP', 'true')

        result = testdir.runpytest('-svv', '-k', 'test_hello_world')
        result.assert_outcomes(passed=1)

        appmap_json = os.path.join(
            str(testdir),
            'tmp', 'appmap', 'pytest', 'test_hello_world.appmap.json'
        )
        with open(appmap_json) as appmap:
            generated_appmap = self.normalize_appmap(appmap.read())

        appmap_json = os.path.join(data_dir, 'pytest', 'expected.appmap.json')
        with open(appmap_json) as f:
            expected_appmap = json.load(f)

        assert generated_appmap == expected_appmap



import importlib.metadata
import logging
import pytest
import os
import os.path
import re

import inflection

from appmap._implementation import configuration, env, generation, recording
from appmap._implementation.utils import fqname

logger = logging.getLogger(__name__)

try:
    # Make sure we have hooks installed if we're testing a django app.
    # pylint: disable=unused-import
    import appmap.django  # noqa: F401
except ImportError:
    pass  # probably no django


class FuncItem:
    def __init__(self, pyfuncitem):
        self.item = pyfuncitem

    @property
    def class_name(self):
        return self.item.cls.__name__

    @property
    def defined_class(self):
        return fqname(self.item.cls)

    @property
    def method_id(self):
        return self.item.originalname

    def is_in_class(self):
        return self.item.cls is not None
    has_feature_group = is_in_class

    @property
    def feature_group(self):
        return inflection.humanize(
            inflection.titleize(self.class_name.removeprefix('Test'))
        )

    @property
    def feature(self):
        return inflection.humanize(self.test_name)

    @property
    def scenario_name(self):
        name = f'{self.feature[0].lower()}{self.feature[1:]}'
        if self.has_feature_group():
            name = ' '.join([self.feature_group, name])
        return name

    @property
    def test_name(self):
        ret = self.item.name
        return re.sub('^test.?_', '', ret)

    @property
    def filename(self):
        fname = self.item.name
        if self.item.cls:
            fname = f'{self.defined_class}_{fname}'
        fname = re.sub('[^a-zA-Z0-9-_]', '_', fname)
        return fname.removesuffix('_')

    @property
    def metadata(self):
        ret = {}
        if self.is_in_class():
            ret['feature_group'] = self.feature_group
            ret['recording'] = {
                'defined_class': self.defined_class,
                'method_id': self.method_id
            }
        ret.update({
            'name': self.scenario_name,
            'feature': self.feature
        })
        return ret


@pytest.hookimpl(hookwrapper=True)
def pytest_runtestloop(session):
    if not env.enabled():
        yield
        return

    pytest_version = importlib.metadata.version('pytest')
    session.appmap_path = os.path.join(
        configuration.Config().output_dir, 'pytest'
    )
    logger.debug('pytest_runtestloop, creating %s', session.appmap_path)
    os.makedirs(session.appmap_path, exist_ok=True)
    session.appmap_metadata = {
        'app': configuration.Config().name,
        'frameworks': [{
            'name': 'pytest',
            'version': pytest_version

        }],
        'recorder': {
            'name': 'pytest'
        }
    }
    yield


@pytest.hookimpl(hookwrapper=True)
def pytest_pyfunc_call(pyfuncitem):
    if not env.enabled():
        yield
        return

    session = pyfuncitem.session
    item = FuncItem(pyfuncitem)
    metadata = dict(session.appmap_metadata)
    metadata.update(item.metadata)
    logger.debug('pytest_pyfunc_call, metadata %s', repr(metadata))
    fname = item.filename + '.appmap.json'
    path = os.path.join(session.appmap_path, fname)

    def write_recording(r):
        with open(path, 'wb') as appmap_file:
            appmap_file.write(generation.dump(r, metadata))
        logger.info('wrote recording %s', path)

    logger.info('starting recording %s', path)
    r = recording.Recording(exit_hook=write_recording)
    with r:
        yield



import importlib
import os
import platform
import re

import json

import appmap._implementation


class AppMapTestBase:
    def setup_method(self, _):
        appmap._implementation.initialize()  # pylint: disable=protected-access

    @staticmethod
    def normalize_git(git):
        git.pop('repository') # = 'git@github.com:applandinc/appmap-python.git'
        git.pop('branch') # = 'master'
        git.pop('commit') # = 'xyz'
        status = git.pop('status')
        assert isinstance(status, list)
        git.pop('git_last_tag') # = ''
        commits_since_last_tag = git.pop('git_commits_since_last_tag')
        assert isinstance(commits_since_last_tag, int)
        git.pop('git_last_annotated_tag') # = None
        commits_since_last_annotated_tag = git.pop(
            'git_commits_since_last_annotated_tag'
        )
        assert isinstance(commits_since_last_annotated_tag, int)

    @staticmethod
    def normalize_metadata(metadata):
        engine = metadata['language'].pop('engine')
        assert engine == platform.python_implementation()
        version = metadata['language'].pop('version')
        assert version == platform.python_version()

        if 'frameworks' in metadata:
            frameworks = metadata['frameworks']
            for f in frameworks:
                if f['name'] == 'pytest':
                    v = f.pop('version')
                    assert v == importlib.metadata.version('pytest')

    def normalize_appmap(self, generated_appmap):
        """
        Normalize the data in generated_appmap, removing any
        environment-specific values.

        Note that attempts to access required keys will raise
        KeyError, causing the test to fail.
        """

        def normalize(dct):
            if 'elapsed' in dct:
                elapsed = dct.pop('elapsed')
                assert isinstance(elapsed, float)
            if 'git' in dct:
                self.normalize_git(dct.pop('git'))
            if 'location' in dct:
                path, line = dct['location'].split(':')
                path = os.path.basename(path)
                dct['location'] = ':'.join([path, line])
            if 'metadata' in dct:
                self.normalize_metadata(dct['metadata'])
            if 'object_id' in dct:
                object_id = dct.pop('object_id')
                assert isinstance(object_id, int)
            if 'path' in dct:
                dct['path'] = os.path.basename(dct['path'])
            if 'value' in dct:
                # This maps all object references to the same
                # location. We don't actually need to verify that the
                # locations are correct -- if they weren't, the
                # instrumented code would be broken, right?
                v = dct['value']
                dct['value'] = re.sub(r'<(.*) object at 0x.*>',
                                      r'<\1 object at 0xabcdef>',
                                      v)
            return dct

        return json.loads(generated_appmap, object_hook=normalize)



from appmap._implementation.event import SqlEvent, ReturnEvent
from appmap._implementation.recording import Recorder

from django.db.backends.signals import connection_created
from django.dispatch import receiver

import time


class ExecuteWrapper:
    def __init__(self):
        self.recorder = Recorder()

    def __call__(self, execute, sql, params, many, context):
        start = time.monotonic()
        try:
            return execute(sql, params, many, context)
        finally:
            if self.recorder.enabled:
                stop = time.monotonic()
                duration = stop - start
                # Note this logic is based on django.db.backends.utils.CursorDebugWrapper.
                if many:
                    # Sometimes the same query is executed with different parameter sets.
                    # Instead of substituting them all, just say how many times it was run.
                    try:
                        times = len(params)
                    except TypeError:
                        times = '?'
                    sql = f'{times} times {sql}'
                else:
                    db = context['connection']
                    cursor = context['cursor']
                    sql = db.ops.last_executed_query(cursor, sql, params)
                call_event = SqlEvent(sql)
                self.recorder.add_event(call_event)
                return_event = ReturnEvent(parent_id=call_event.id, elapsed=duration)
                self.recorder.add_event(return_event)


@receiver(connection_created)
def connected(sender, connection, **_):
    # pylint: disable=unused-argument
    wrappers = connection.execute_wrappers
    if not any(isinstance(x, ExecuteWrapper) for x in wrappers):
        wrappers.append(ExecuteWrapper())



import inspect
import threading
from itertools import chain
from functools import partial

from .utils import appmap_tls, split_function_name, fqname


class _EventIds:
    id = 1
    lock = threading.Lock()

    @classmethod
    def next_id(cls):
        with cls.lock:
            cls.id += 1
            return cls.id

    # The identifiers returned by threading.get_ident() aren't
    # guaranteed to be unique: they may be recycled after the thread
    # exits. We need a unique id, so we'll manage it ourselves.
    _next_thread_id = 0
    _next_thread_id_lock = threading.Lock()

    @classmethod
    def next_thread_id(cls):
        with cls._next_thread_id_lock:
            cls._next_thread_id += 1
            return cls._next_thread_id

    @classmethod
    def get_thread_id(cls):
        tls = appmap_tls()
        return (tls.get('thread_id')
                or tls.setdefault('thread_id', cls.next_thread_id()))

    @classmethod
    def reset(cls):
        cls.id = 1
        cls._next_thread_id = 0


class Event:
    __slots__ = ['id', 'event', 'thread_id']

    def __init__(self, event):
        self.id = _EventIds.next_id()
        self.event = event
        self.thread_id = _EventIds.get_thread_id()

    def to_dict(self):
        return {
            k: getattr(self, k, None)
            for k in chain.from_iterable(getattr(cls, '__slots__', [])
                                         for cls in type(self).__mro__)
        }

    def __repr__(self):
        return repr(self.to_dict())


class CallEvent(Event):
    __slots__ = ['defined_class', 'method_id', 'path', 'lineno',
                 'static', 'receiver', 'parameters']

    @staticmethod
    def make(fn, isstatic):
        """
        Return a factory for creating new CallEvents based on
        introspecting the given function.
        """
        defined_class, method_id = split_function_name(fn)

        try:
            path = inspect.getsourcefile(fn)
        except TypeError:
            path = '<builtin>'

        try:
            __, lineno = inspect.getsourcelines(fn)
        except OSError:
            lineno = 0

        return partial(CallEvent, defined_class,
                       method_id, path, lineno, isstatic)

    @staticmethod
    def make_receiver(fn, isstatic):
        """
        Create the receiver object that should be part of the call
        event for the given function.
        """
        defined_class, __ = split_function_name(fn)
        if isstatic:
            cls = "class"
            value = defined_class
            object_id = id(defined_class)
        else:
            cls = defined_class
            value = None
            object_id = None

        def make(cls, value, object_id, *args):
            if not isstatic:
                object_id = id(args[0][0])
                slf = args[0][0]
                # Make a best-effort attempt to get a string value for
                # the receiver. If repr() raises, formulate a
                # value from the class and id.
                try:
                    value = repr(slf)
                except Exception:  # pylint: disable=broad-except
                    value = f'<{defined_class} object at {object_id:#02x}>'
            return {
                "class": cls,
                "value": value,
                "object_id": object_id
            }
        return partial(make, cls, value, object_id)

    def __init__(self, defined_class, method_id, path, lineno,
                 static, receiver, parameters):
        super().__init__('call')
        self.defined_class = defined_class
        self.method_id = method_id
        self.path = path
        self.lineno = lineno
        self.static = static
        self.receiver = receiver
        self.parameters = parameters


class SqlEvent(Event):
    __slots__ = ['sql_query']

    def __init__(self, sql):
        super().__init__('call')
        self.sql_query = {
            'sql': sql
        }


class ReturnEvent(Event):
    __slots__ = ['parent_id', 'elapsed']

    def __init__(self, parent_id, elapsed):
        super().__init__('return')
        self.parent_id = parent_id
        self.elapsed = elapsed


class ExceptionEvent(ReturnEvent):
    __slots__ = ['exceptions']

    def __init__(self, parent_id, elapsed, exc_info):
        super().__init__(parent_id, elapsed)
        class_, exc, __ = exc_info
        self.exceptions = [{
            'exceptions': {
                'class': fqname(class_),
                'message': str(exc),
                'object_id': id(exc),
            }
        }]


def initialize():
    appmap_tls().pop('thread_id', None)
    _EventIds.reset()



import threading
import types
from collections.abc import MutableMapping
import subprocess


class ThreadLocalDict(threading.local, MutableMapping):
    def __init__(self):
        super().__init__()
        self.values = {}

    def __getitem__(self, k):
        return self.values[k]

    def __setitem__(self, k, v):
        self.values[k] = v

    def __delitem__(self, k):
        del self.values[k]

    def __iter__(self):
        return iter(self.values)

    def __len__(self):
        return len(self.values)


_appmap_tls = ThreadLocalDict()


def appmap_tls():
    return _appmap_tls


def is_staticmethod(m):
    return isinstance(m, (staticmethod, types.BuiltinMethodType))


def is_classmethod(m):
    return isinstance(m, (classmethod, types.BuiltinMethodType))


def fqname(cls):
    return f'{cls.__module__}.{cls.__qualname__}'


def split_function_name(fn):
    """
    Given a method, return a tuple containing its fully-qualified
    class name and the method name.
    """
    qualname = fn.__qualname__
    if '.' in qualname:
        class_name, fn_name = qualname.rsplit('.', 1)
        class_name = f'{fn.__module__}.{class_name}'
    else:
        class_name = fn.__module__
        fn_name = qualname
    return (class_name, fn_name)


def subprocess_run(command_args):
    return subprocess.run(command_args, capture_output=True, text=True, check=False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['appmap',
 'appmap._implementation',
 'appmap.test',
 'appmap.test.data',
 'appmap.test.data.package1',
 'appmap.test.data.package1.package2',
 'appmap.test.data.pytest']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0', 'inflection>=0.5.1,<0.6.0', 'orjson==3.4.6']

entry_points = \
{'pytest11': ['appmap = appmap.pytest']}

setup_kwargs = {
    'name': 'appmap',
    'version': '0.1.0.dev7',
    'description': 'Create AppMap files by recording a Python application.',
    'long_description': '## About\n`appmap-python` is a Python package for recording\n[AppMaps](https://github.com/applandinc/appmap) of your code. "AppMap" is a data format\nwhich records code structure (modules, classes, and methods), code execution events\n(function calls and returns), and code metadata (repo name, repo URL, commit SHA, labels,\netc). It\'s more granular than a performance profile, but it\'s less granular than a full\ndebug trace. It\'s designed to be optimal for understanding the design intent and structure\nof code and key data flows.\n\nThere are several ways to record AppMaps of your Python program using the `appmap` package:\n\n* Run your tests (pytest, unittest[future work]) with the environment variable\n  `APPMAP=true`. An AppMap will be generated for each test case.\n    \n* Use the `appmap.record` [context manager](#context-manager) to control recording. The context manager takes\n  an instance of an `appmap.Recording`, which can be used to generate the AppMap.\n\n* [future work] Run your application server with AppMap remote recording enabled, and use\n  the [AppLand browser extension](https://github.com/applandinc/appland-browser-extension)\n  to start, stop, and upload recordings.\n  \nOnce you have made a recording, there are two ways to view automatically generated\ndiagrams of the AppMaps.\n\nThe first option is to load the diagrams directly in your IDE, using the [AppMap extension\nfor VSCode](https://marketplace.visualstudio.com/items?itemName=appland.appmap).\n\nThe second option is to upload them to the [AppLand server](https://app.land) using the\n[AppLand CLI](https://github.com/applandinc/appland-cli/releases).  \n\n### Supported versions\n\n* Python 3.9\n* Pytest 6.2\n\nSupport for new versions is added frequently, please check back regularly for updates.\n\n## Configuration\nAdd your modules as `path` entries in `appmap.yml`:\n\n```yaml\nname: my_python_app\npackages:\n- path: app.mod1\n- path: app.mod2\n  exclude:\n  - MyClass\n  - MyOtherClass.my_instance_method\n  - MyOtherClass.my_class_method\n```\n\nNote that `exclude`s are resolved relative to the associated path. So, for example, this\nconfiguration excludes `app.mod2.MyClass`\n\n## Environment Variables\n\n* `APPMAP` if `true`, code will be instrumented and AppMaps will be generated. Not\n  case-sensitive, defaults to \'false\'.\n\n* `APPMAP_CONFIG` specifies the configuration file to use. Defaults to `appmap.yml` in the\n  current directory\n  \n* `APPMAP_LOG_LEVEL` specifies log level to use, from the set `CRITICAL`, `ERROR`,\n  `WARNING`, `INFO`, `DEBUG`. Not case-sensitive, defaults to `WARNING`.\n  \n* `APPMAP_OUTPUT_DIR` specifies the root directory for writing AppMaps. Defaults to\n  `tmp/appmap`.\n  \n## Test Frameworks\n`appmap-python` supports recording [pytest](https://pytest.org) and `unittest`[future\nwork] test cases.\n\n### pytest\n`appmap-python` is a `pytest` plugin. When it\'s installed in a project that uses\n`pytest`, it will be available to generate AppMaps.\n\n```shell\nroot@e9987eaa93c8:/src/appmap/test/data/pytest# pip show appmap\nName: appmap\nVersion: 0.0.0\nSummary: Create AppMap files by recording a Python application.\nHome-page: None\nAuthor: Alan Potter\nAuthor-email: alan@app.land\nLicense: None\nLocation: /usr/local/lib/python3.9/site-packages\nRequires: orjson, PyYAML, inflection\nRequired-by:\nroot@e9987eaa93c8:/src/appmap/test/data/pytest# APPMAP=true APPMAP_LOG_LEVEL=info pytest -svv\n[2021-02-10 11:37:59,345] INFO root: appmap enabled: True\n[2021-02-10 11:37:59,350] INFO appmap._implementation.configuration: ConfigFilter, includes {\'simple\'}\n[2021-02-10 11:37:59,350] INFO appmap._implementation.configuration: ConfigFilter, excludes set()\n===================================================================== test session starts =====================================================================\nplatform linux -- Python 3.9.1, pytest-6.2.2, py-1.10.0, pluggy-0.13.1 -- /usr/local/bin/python\ncachedir: .pytest_cache\nrootdir: /src, configfile: pytest.ini\nplugins: appmap-0.0.0\ncollected 1 item\n\ntest_simple.py::test_hello_world [2021-02-10 11:37:59,482] INFO appmap.pytest: starting recording /tmp/pytest/test_hello_world.appmap.json\n[2021-02-10 11:37:59,484] INFO appmap._implementation.configuration: included class simple.Simple\n[2021-02-10 11:37:59,484] INFO appmap._implementation.configuration: included function simple.Simple.hello\n[2021-02-10 11:37:59,489] INFO appmap._implementation.configuration: included function simple.Simple.hello_world\n[2021-02-10 11:37:59,490] INFO appmap._implementation.configuration: included function simple.Simple.world\n[2021-02-10 11:37:59,828] INFO appmap.pytest: wrote recording /tmp/pytest/test_hello_world.appmap.json\nPASSED\n\n====================================================================== 1 passed in 0.45s ======================================================================\n```\n\n### unittest [future work]\n`import appmap.unittest`. Instruments subclasses of `unittest.TestCase` and records each\n`test_*` function in the subclasses.\n\n### Run your tests\nOnce you\'ve configured your tests to generate AppMaps, run the tests with the\n`APPMAP=true` in the environment. For example, to run a pytest test suite:\n\n```sh\n$ APPMAP=true pytest\n```\n\n\n## Remote Recording [future work]\n`appmap-python` supports remote recording of Django and Flask web applications. Import the\nappropriate remote recording support into your web app.\n\n### Django\n`import appmap.django`. Adds `/_appmap/record` routes to a Django app.\n\n## Flask\n`import appmap.flask`. Adds `/_appmap/record` routes to a Flask app.\n\n## Run your web app [future work]\nOnce you\'ve configured your web app to add the remote-recording routes, you can use the\nroutes to manage recordings. The browser extension, appland CLI, or just plain cURL will\nall work for this.\n\nAs when running tests, start the web server with `APPMAP=true` in the environment. For\nexample, to start a Django app:\n\n```sh\n$ APPMAP=true python manage.py runserver\n```\n\nAn app with remote recording enabled supports these routes:\n\n* `POST /_appmap/record`\n  Starts a new recording\n\n  200 if the recording was started successfully\n  409 if there\'s already a recording in progress\n\n* `GET /_appmap/record`\n  Returns JSON describing current recording state\n  200 with body\n\n  ```json\n  {\n    "enabled": true\n  }\n  ```\n  `enabled` indicates whether recording has been enabled\n\n* `DELETE /_appmap/record`\n  Returns AppMap as JSON\n  200 with AppMap as body\n  404 if there\'s no recording in progress\n\n## Context manager\nYou can use `appmap.record` as a context manager to record your code.\n\nWith a file called `record_sample.py` like this\n\n```python\nimport os\nimport sys\n\nimport appmap\n\nr = appmap.Recording()\nwith r:\n    import sample\n    print(sample.C().hello_world(), file=sys.stderr)\n\nwith os.fdopen(sys.stdout.fileno(), "wb", closefd=False) as stdout:\n    stdout.write(appmap.generation.dump(r))\n    stdout.flush()\n```\n\nand a source file called `sample.py` like this\n\n```python\nclass C:\n    def make_str(self, s):\n        return s;\n\n    def hello_world(self):\n        return f\'{self.make_str("Hello")} {self.make_str("world!")}\'\n```\n\nas well as an `appmap.yml`\n\n```yaml\nname: sample\npackages:\n- path: sample\n```\n\nyou can generate a recording of the code\n\n```sh\n% APPMAP=true python record_sample.py > record_sample.appmap.json\n% jq \'.events | length\' record_sample.appmap.json\n6\n% jq < record_sample.appmap.json | head -10\n{\n  "version": "1.4",\n  "metadata": {\n    "language": {\n      "name": "python",\n      "engine": "CPython",\n      "version": "3.9.1"\n    },\n    "client": {\n      "name": "appmap",\n```\n\n## Development\n\n[![Build Status](https://travis-ci.com/applandinc/appmap-python.svg?branch=master)](https://travis-ci.com/applandinc/appmap-python)\n\n### Dependency management\n\n[poetry](https://https://python-poetry.org/) for dependency management:\n\n```\n% brew install poetry\n% cd appmap-python\n% poetry install\n```\n\n### Linting\n[pylint](https://www.pylint.org/) for linting:\n\n```\n% cd appmap-python\n% poetry run pylint appmap\n\n--------------------------------------------------------------------\nYour code has been rated at 10.00/10 (previous run: 10.00/10, +0.00)\n\n```\n\n[Note that the current configuration requires a 10.0 for the Travis build to pass. To make\nthis easier to achieve, convention and refactoring checks have both been disabled. They\nshould be reenabled as soon as possible.]\n\n\n### Testing\n[pytest](https://docs.pytest.org/en/stable/) for testing:\n\n```\n% cd appmap-python\n% APPMAP=true poetry run pytest\n```\n\n### Code Coverage\n[coverage](https://coverage.readthedocs.io/) for coverage:\n\n```\n% cd appmap-python\n% APPMAP=true poetry run coverage run -m pytest\n% poetry run coverage html\n% open htmlcov/index.html\n```\n',
    'author': 'Alan Potter',
    'author_email': 'alan@app.land',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Test Configuration"""
import os

import appmap
from .appmap_test_base import AppMapTestBase
from appmap._implementation.configuration import ConfigFilter
from appmap._implementation.recording import NullFilter


class TestConfiguration(AppMapTestBase):
    def test_can_be_enabled(self, monkeypatch):
        """Test that recording is enabled when APPMAP=true"""
        monkeypatch.setenv("APPMAP", "true")
        assert appmap.enabled()

    def test_is_disabled_when_unset(self, monkeypatch):
        """Test that recording is disabled when APPMAP is unset"""
        monkeypatch.delenv("APPMAP", raising=False)
        assert not appmap.enabled()

    def test_is_disabled_when_false(self, monkeypatch):
        """Test that recording is disabled when APPMAP=false"""
        monkeypatch.setenv("APPMAP", "false")
        assert not appmap.enabled()

    def test_package_included(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1'
        c.__qualname__ = 'cls'
        assert f.filter(c) is True

    def test_class_included(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap-class.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1.package2'
        c.__qualname__ = 'Mod1Class'
        assert f.filter(c) is True

    def test_function_included(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap-func.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1.package2'
        c.__qualname__ = 'Mod1Class.func'
        assert f.filter(c) is True

    def test_function_included_by_package(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1.package2'
        c.__qualname__ = 'Mod1Class.func'
        assert f.filter(c) is True

    def test_function_included_by_class(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap-class.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1.package2'
        c.__qualname__ = 'Mod1Class.func'
        assert f.filter(c) is True

    def test_class_prefix_doesnt_match(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = ConfigFilter(NullFilter())
        c = mocker.Mock()
        c.__module__ = 'package1_prefix'
        c.__qualname__ = 'cls'
        assert f.filter(c) is False

    def test_fn_prefix_doesnt_match(self, data_dir, mocker, monkeypatch):
        monkeypatch.setenv("APPMAP", "true")
        monkeypatch.setenv("APPMAP_CONFIG",
                           os.path.join(data_dir, 'appmap.yml'))
        monkeypatch.setenv("APPMAP_LOG_LEVEL", 'debug')

        f = mocker.Mock()
        f.__module__ = 'package1_prefix'
        f.__qualname__ = 'cls.func'
        fltr = ConfigFilter(NullFilter())
        assert not fltr.included(f)



"""
Manage Configuration AppMap recorder for Python.
"""

import logging
import os
import sys
from contextlib import contextmanager
from functools import wraps, cached_property
import time

import yaml

from . import env
from . import event
from .recording import Recorder, Filter
from .utils import appmap_tls, split_function_name, fqname

logger = logging.getLogger(__name__)


class Config:
    """ Singleton Config class """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            logger.debug('Creating the Config object')
            cls._instance = super(Config, cls).__new__(cls)

            # If we're not enabled, no more initialization needs to be
            # done.
            cls._instance._initialized = not env.enabled()

        return cls._instance

    def __init__(self):
        if self.__getattribute__('_initialized'):  # keep pylint happy
            return

        config_file = os.getenv("APPMAP_CONFIG", "appmap.yml")
        with open(config_file) as file:
            self._config = yaml.load(file, Loader=yaml.BaseLoader)
            logger.debug('self._config %s', self._config)

        self._initialized = True

    @classmethod
    def initialize(cls):
        cls._instance = None

    @cached_property
    def output_dir(self):
        return os.getenv("APPMAP_OUTPUT_DIR",
                         os.path.join('tmp', 'appmap'))

    @cached_property
    def name(self):
        return self._config['name']

    @cached_property
    def packages(self):
        return self._config['packages']


@contextmanager
def recording_disabled():
    tls = appmap_tls()
    tls['instrumentation_disabled'] = True
    try:
        yield
    finally:
        tls['instrumentation_disabled'] = False


def is_instrumentation_disabled():
    return appmap_tls().setdefault('instrumentation_disabled', False)


def wrap(fn, isstatic):
    """return a wrapped function"""
    logger.debug('hooking %s', '.'.join(split_function_name(fn)))

    make_call_event = event.CallEvent.make(fn, isstatic)
    make_receiver = event.CallEvent.make_receiver(fn, isstatic)

    @wraps(fn)
    def wrapped_fn(*args, **kwargs):
        if not Recorder().enabled or is_instrumentation_disabled():
            return fn(*args, **kwargs)

        with recording_disabled():
            logger.debug('%s args %s kwargs %s', fn, args, kwargs)
            call_event = make_call_event(receiver=make_receiver(args, kwargs),
                                         parameters=[])
        call_event_id = call_event.id
        Recorder().add_event(call_event)
        start_time = time.time()
        try:
            ret = fn(*args, **kwargs)
            elapsed_time = time.time() - start_time

            return_event = event.ReturnEvent(parent_id=call_event_id, elapsed=elapsed_time)
            Recorder().add_event(return_event)
            return ret
        except Exception:  # noqa: E722
            elapsed_time = time.time() - start_time
            Recorder().add_event(event.ExceptionEvent(parent_id=call_event_id,
                                                      elapsed=elapsed_time,
                                                      exc_info=sys.exc_info()))
            raise
    setattr(wrapped_fn, '_appmap_wrapped', True)
    return wrapped_fn


def name_in_set(name, which):
    return any(filter(lambda n: name.startswith(n + '.') or name == n,
                      which))


def class_in_set(class_name, which):
    return name_in_set(class_name + '.', which)


def function_in_set(fn, which):
    class_name, fn_name = split_function_name(fn)
    if class_name is None:
        # fn isn't in a class, ignore it.
        return False

    ret = (class_in_set(class_name, which)
           or name_in_set(f'{class_name}.{fn_name}', which))

    logger.debug(('function_in_set, class_name %s'
                  ' fnname %s'
                  ' which %s'
                  ' ret %s'),
                 class_name,
                 fn_name,
                 which,
                 ret)
    return ret


class ConfigFilter(Filter):
    """
    During initialization, the ConfigFilter class reads the `package`
    entries from the config file.  Based on these entries, it creates
    a set of paths that should be included, and another set of paths
    that should be excluded.

    When checking to see if a function should be included,
    ConfigFilter looks for the function in both sets.  If the function
    is found in the included set, it's wrapped for instrumentation.
    If it's found in the excluded set, it's used without modification.
    If it isn't found in either set, ConfigFilter passes the function
    on to the next Filter.

    To determine whether a function is in one of the sets, the
    fully-qualified name of the function is considered, including the
    packages(s), module, and class that contain it.  As an example,
    consider a function named `pkg1.pkg2.mod1.Class1.func`.  Any of
    `pkg1`, `pkg1.pkg2.`, `pkg1.pkg2.mod1`, `pkg1.pkg2.mod1.Class1`,
    or `pkg1.pkg2.mod1.Class1.func` will match this function (causing
    it to be included or excluded, as appropriate).  Note that only
    full names will match.  (This is achieved in the code by appending
    `.` to the name retrieved from the config file.)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._includes = set()
        self._excludes = set()

        if not env.enabled():
            return

        for package in Config().packages:
            path = package['path']
            self._includes.add(path)
            if 'exclude' in package:
                if not isinstance(package['exclude'], list):
                    raise RuntimeError('Excludes for package'
                                       f' "{path}" must be a list')
                excludes = [f'{path}.{e}' for e in package['exclude']]
                self._excludes.update(excludes)
        logger.info('ConfigFilter, includes %s', self._includes)
        logger.info('ConfigFilter, excludes %s', self._excludes)

    def excluded(self, fn):
        ret = function_in_set(fn, self._excludes)
        logger.debug('ConfigFilter, %s excluded? %s', fn, ret)
        return ret

    def included(self, fn):
        ret = function_in_set(fn, self._includes)
        logger.debug('ConfigFilter, %s included? %s', fn, ret)
        return ret

    def filter(self, class_):
        name = fqname(class_)
        if class_in_set(name, self._excludes):
            logger.info('excluded class %s', name)
            return False
        if class_in_set(name, self._includes):
            logger.info('included class %s', name)
            return True

        logger.debug('  undecided')
        return self.next_filter.filter(class_)

    def wrap(self, fn, isstatic):
        logger.debug('ConfigFilter.wrap, fn %s', fn)

        fn_name = '.'.join(split_function_name(fn))
        if self.excluded(fn):
            logger.info('excluded function %s', fn_name)
            return fn

        if self.included(fn):
            logger.info('included function %s', fn_name)
            wrapped = getattr(fn, '_appmap_wrapped', None)
            logger.debug('  wrapped %s', wrapped)
            if wrapped is None:
                logger.debug('  wrapping %s', fn)
                ret = wrap(fn, isstatic)
            else:
                logger.debug('  already wrapped %s', fn)
                ret = fn
            return ret

        return self.next_filter.wrap(fn, isstatic)


class BuiltinFilter(Filter):
    def __init__(self, *args):
        super().__init__(*args)
        if not env.enabled():
            self._includes = set()
            return

        self._includes = {'os.read', 'os.write'}

    def included(self, fn):
        return function_in_set(fn, self._includes)

    def filter(self, class_):
        name = class_.__name__
        if class_in_set(name, self._includes):
            return True
        return self.next_filter.filter(class_)

    def wrap(self, fn, isstatic):
        if self.included(fn):
            return wrap(fn, isstatic)
        return self.next_filter.wrap(fn, isstatic)


def initialize():
    Config().initialize()
    Recorder().use_filter(BuiltinFilter)
    Recorder().use_filter(ConfigFilter)


initialize()
